#include "RRFixPointLib.h"


#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>


//--------------------------------------------------------------------------------------
//--- Basis-Fixpoint-Arithmetik-Routinen ------------------- Version vom 29.Dez.2007 ---
//--- Zulssiger Wertebereich ist -32767.0xx bis 32767.0xx -----------------------------
//--- Die Nachkommagenauigkeit betrgt lediglich 10 Bit! -------------------------------
//--- atofixpoint und fixpointtoa wandeln so gut sie knnen, aber nicht przise --------
//--- Also nicht vergessen: Die Rechenergebnisse sind reichlich UNGENAU ----------------
//--------------------------------------------------------------------------------------
//--- (c) 2007 Ralf Rosenkranz ---------------------------------------------------------
//--- voltax tt prima dott de ---------------------------------------------------------
//--- http://basic.3dc.de --------------------------------------------------------------
//--------------------------------------------------------------------------------------
//--- Die Routinen scheinen zu funktionieren aber ich garantiere fr nichts ------------
//--- Ich hafte fr keinerlei Schden und bernehme keine Gewhrleistung ---------------
//--- Dieser Quellcode darf nur fr *nichtkommerzielle* Zwecke frei verwendet werden ---
//-------------------------------------------------------------------------------------- 



#define pixpointFractShift 10 //--- Hier nichts anderes als 10 !!! ---
#define fixpointBits 32
#define fixpointRealBits (fixpointBits-1-pixpointFractShift)


fixpoint int16ToFixpoint (int16_t i)
{
   return ((fixpoint) i) << pixpointFractShift;
}   


int16_t fixpointToInt16 (fixpoint f)
{
   return (int16_t) (f >> pixpointFractShift);   
}   



fixpoint atofixpoint (char* str)
{
   char vkBuf [6];
   char nkBuf [4];

   bool neg = false;
   uint8_t n = 0;
   uint8_t m = 0;
   uint8_t c;
   bool pointFound = false;
   
   
   while (((c = str [n]) != '.') && (c != 0) && (n < sizeof (vkBuf) - 1))
   {
      if ((n == 0) && (c == '-'))
      {
         neg = true;
      }
      else
      {
         vkBuf [m] = c;
         m++;
      }

      n++;
   }

   vkBuf [m] = 0;
   m = 0;

   if (c == '.')
   { 
      pointFound = true;
      n++;

      while (((c = str [n]) != 0) && (m < sizeof (nkBuf) - 1))
      {
         nkBuf [m] = c;
         n++;
         m++;
      }

      nkBuf [m] = 0;
   }
   else
   {
      nkBuf [0] = 0;
   }
         
   
   int32_t i = atoi (vkBuf);
   fixpoint f = ((fixpoint) i) << pixpointFractShift;
   
   
   if (pointFound)
   {
      for (m = 0; m < sizeof (nkBuf) - 1; m++)  
      {
         if (nkBuf [m] == 0)
         {
            nkBuf [m] = '0';
            nkBuf [m+1] = 0;
         }
      }
   
   
      int16_t r = atoi (nkBuf);            
      int16_t sr = r;
   
   
      if (sr != 0)
      {
         if (sr < 0)
         {
            sr = -sr;
         }
         
         sr = (sr*17)/16; //--- Scale 500 -> 512 ---
         sr = (sr*28)/29; //--- nicht genau aber ok ---
         
         if (sr > 1023)      
         {
            sr = 1023;
         }
         else if (sr < 1)      
         {
            sr = 1;
         }
   
         f |= sr;
      }
   }   

   if (neg)
   {
      return -f;
   }
   else
   {
      return f;
   }
}   



char* fixpointtoa (fixpoint f, char* str)
{
   char* result = str;

   if (f < 0)
   {
      f = - f;
      *str++ = '-';
   }

   uint16_t vk = f >> pixpointFractShift;
   uint16_t nk = f & ((1 << pixpointFractShift) - 1);
   uint16_t snk;

   itoa (vk, str, 10);
   
   if (nk != 0)
   {
      snk = (nk*29)/28;  //--- Scale 512 -> 500 ---
      snk = (snk*17)/18; //--- nicht genau aber ok ---
      
      if (snk > 999)      
      {
         snk = 999;
      }
      else if (snk < 1)
      {      
         snk = 1;
      }

      size_t len = strlen (str);
      str [len++] = '.';

      uint16_t p = nk * 10;
      uint16_t t = 1024;
   
      while (p < t)
      {
         str [len++] = '0';
         p *= 10;
      }

      itoa (snk, str + len, 10);   
   }
   
   return result;
}      



fixpoint fixpointAdd (fixpoint a, fixpoint b)
{
   return a + b;   
}



fixpoint fixpointSub (fixpoint a, fixpoint b)
{
   return a - b;
}



fixpoint fixpointMult (fixpoint a, fixpoint b)
{   
   fixpoint res;
   bool resNeg = false;
   
   uint8_t aBits = 0;
   fixpoint aProbe;
  
   if (a < 0)
   {
      a = -a;
      resNeg = !resNeg;
   }
 
   aProbe = a;

   while (aProbe != 0)
   {
      aProbe >>= 1;
      aBits ++;
   }
  
   uint8_t bBits = 0;
   fixpoint bProbe;

   if (b < 0)
   {
      b = -b;
      resNeg = !resNeg;
   }

   bProbe = b;      

   while (bProbe != 0)
   {
      bProbe >>= 1;
      bBits ++;
   }
  
   if (aBits + bBits < fixpointBits)
   {
      res = (a * b) >> pixpointFractShift;      
   }
   else
   {
      uint8_t resShift = 0;

      while (aBits + bBits >= fixpointBits)
      {
         if (aBits >= bBits)
         {
            a >>= 1;            
            aBits --;            
         }
         else
         {

            b >>= 1;            
            bBits --;            
         }

         resShift ++;
      }

      if (pixpointFractShift >= resShift)
      {
         res = (a * b) >> (pixpointFractShift - resShift);
      }
      else
      {
         res = (a * b) << (resShift - pixpointFractShift);
      }
   }

   if (resNeg)
   {
      return -res;
   }
   else
   {
      return res;
   }
}



fixpoint fixpointDiv (fixpoint a, fixpoint b)
{ 
   fixpoint res;
   bool resNeg = false;
   
   uint8_t aBits = 0;
   fixpoint aProbe;
  
   if (a < 0)
   {
      a = -a;
      resNeg = !resNeg;
   }
   
   aProbe = a;
      
   if (b < 0)
   {
      b = -b;
      resNeg = !resNeg;
   }

   while (aProbe != 0)
   {
      aProbe >>= 1;
      aBits ++;
   }
    
   if (aBits < fixpointRealBits)
   {    
      res = (a << pixpointFractShift) / b;
   }
   else
   {
      uint8_t resShift = 0;

      while (aBits >= fixpointRealBits)
      {
         a >>= 1;            
         aBits --;                    
         resShift ++;
      }

      res = ((a << pixpointFractShift) / b) << resShift;      
   }
   
   if (resNeg)
   {
      return -res;
   }
   else
   {
      return res;
   }   
}



fixpoint fixpointSqr (fixpoint x)
{   
   return fixpointMult (x, x);
}



fixpoint fixpointSqrt (fixpoint x)
{
   uint32_t r = 0;         
   uint32_t hi = 0;       
   uint32_t lo = x;        
   uint32_t c = 15 + (pixpointFractShift >> 1);
   uint32_t d;

   do 
   {
      hi = (hi << 2) | (lo >> 30);  
      lo <<= 2;
      r <<= 1;   
      d = (r << 1) + 1;   
      
      if (hi >= d) 
      {
         hi -= d;
         r += 1;
      }
      
   } while (c-- != 0);

   return (r);
}



